# coding:utf-8

require 'zlib'
require 'rexml/document'

module Monpe

  #不正なフォーマットを示す例外
  class FormatError < StandardError; end

  #パーサーエラーを示す例外
  class ParserError < StandardError; end

  #帳票ファイルのパースを行う。
  #クラスメソッドのみ。
  class Parser
   
    class << self

      #:call-seq:
      #  parser(filename) -> Hash
      #
      #帳票ファイルのパースを行い、結果をHashオブジェクトとして返す
      #
      #* 引数 filename: パースするファイルのパス
      #* 戻り値: パース結果(Hash)
      def parse(filename)
        info = {
          "dictionarydata" => {},
          "diagramdata"    => {},
          "layers"         => []
        }
        red_data = read_red(filename)
        begin
          doc = REXML::Document.new(red_data)
        rescue Exception => ex
          raise FormatError.new "invalid format:#{filename} #{ex}"
        end
  
        elements =  doc.get_elements("dia:diagram/dia:dictionarydata/dia:element")
        elements << doc.get_elements("dia:diagram/dia:dictionarydata/element")
        elements.flatten!
        elements.each do |elem|
          n,v = parse_dic_element(elem)
          info["dictionarydata"][n] = v
        end
  
        doc.get_elements("dia:diagram/dia:diagramdata/dia:attribute").each do |attr|
          k,v,t = parse_attribute(attr)
          info["diagramdata"][k] =  {"val"=>v,"type"=>t} if k
        end
        doc.get_elements("dia:diagram/dia:layer").each do |layer|
          info["layers"] << parse_layer(layer)
        end
        raise ParserError.new("no layers")      if info["layers"].empty?
        raise ParserError.new("no diagramdata") if info["diagramdata"].empty?
        info
      end

      private

      def read_red(filename)
        begin
          out = ""
          Zlib::GzipReader.open(filename) do |gz|
            out = gz.read
          end
        rescue Zlib::GzipFile::Error
          out = File.read(filename)
        end
        out
      end

      def parse_composite(elem)
        ret = {"type" => elem.attribute("type").to_s, "attributes" => {}}
        elem.get_elements("dia:attribute").each do |a|
          k,v = parse_attribute(a)
          ret["attributes"][k] = v if k
        end
        ret
      end

      def parse_attribute(elem)
        name = elem.attribute("name")
        return nil,nil,nil unless name
        name = name.to_s
        val = nil
        case name
        # enum
        when "alignment", "aspect", "embed_char_type", "embed_text_string",
             "start_arrow", "end_arrow", "line_style", "valign"
          child = elem.get_elements("dia:enum")[0]
          val = child.attribute("val").to_s.to_i
        # color
        when "background","border_color","color","line_color",
             "fill_color","inner_color","pagebreak","arc_color"
          child = elem.get_elements("dia:color")[0]
          if /#([a-f\d]{2})([a-f\d]{2})([a-f\d]{2})/=~ child.attribute("val").to_s
            val = Monpe::Color.new(
              (($1.to_i(16)) * 1.0 / 255.0),
              (($2.to_i(16)) * 1.0 / 255.0),
              (($3.to_i(16)) * 1.0 / 255.0))
          else
            val = Monpe::Color.new(0,0,0)
          end
        # real
        when "bmargin","border_width","corner_radius","curve_distance",
             "custom_height","custom_width","dashlength","elem_height",
             "elem_width","start_arrow_length","start_arrow_width",
             "end_arrow_length","end_arrow_width","height",
             "line_width","lmargin","rmargin","tmargin","multi","single",
             "psheight","pswidth","scaling","width_x","width_y",
             "absolute_start_gap","absolute_end_gap"
          child = elem.get_elements("dia:real")[0]
          val = child.attribute("val").to_s.to_f
        when "poly_points"
          val = []
          elem.get_elements("dia:point").each do |c|
            %r|^([\d\.]*),([\d\.]*)$| =~ c.attribute("val").to_s
            val << Monpe::Point.new($1.to_f,$2.to_f)
          end
        # point x 2
        when "conn_endpoints"
          val = []
          elem.get_elements("dia:point").each do |c|
            %r|^([\d\.]*),([\d\.]*)$| =~ c.attribute("val").to_s
            val << Monpe::Point.new($1.to_f,$2.to_f)
          end
        # string
        when "dnode_path","embed_id","file","name","string"
          child = elem.get_elements("dia:string")[0]
          %r|\A#(.*)#\z|m =~ child.text
          val = $1
        # boolean
        when "draw_border","fitto","is_portrait","keep_aspect","keep_orig_aspect",
             "show_background"
          child = elem.get_elements("dia:boolean")[0]
          val = child.attribute("val").to_s == "true"
        # point
        when "elem_corner","obj_pos","pos"
          child = elem.get_elements("dia:point")[0]
          %r|^([\d\.]*),([\d\.]*)$| =~ child.attribute("val").to_s
          val = Monpe::Point.new($1.to_f,$2.to_f)
        # int
        when "embed_column_size","embed_path_size","embed_text_column","embed_text_size",
             "numcp","visible_x","visible_y"
          child = elem.get_elements("dia:int")[0]
          val = child.attribute("val").to_s.to_i
        # font
        when "font"
          child = elem.get_elements("dia:font")[0]
          family = child.attribute("family").to_s
          family = 'TakaoMincho'   if family == 'Takao明朝'
          family = 'TakaoExMincho' if family == 'TakaoEx明朝'
          family = 'TakaoPMincho'  if family == 'TakaoP明朝'
          family = 'TakaoGothic'   if family == 'Takaoゴシック'
          family = 'TakaoEXGothic' if family == 'TakaoExゴシック'
          family = 'TakaoPGothic'  if family == 'TakaoPゴシック'
          val = Monpe::Font.new(
            family,
            child.attribute("style").to_s.to_i)
        # rectangle
        when "obj_bb"
          child = elem.get_elements("dia:rectangle")[0]
          %r|^([\d\.]*),([\d\.]*);([\d\.]*),([\d\.]*)$| =~ child.attribute("val").to_s
          val = Monpe::Rectangle.new($1.to_f, $2.to_f, $3.to_f, $4.to_f)
        when "text"
          composite = elem.get_elements("dia:composite")[0]
          v = parse_composite(composite)
          attrs  = v["attributes"]
          str    = attrs["string"]
          font   = attrs["font"]
          height = attrs["height"]
          pos    = attrs["pos"]
          color  = attrs["color"]
          align  = attrs["alignment"]
          val = Monpe::Text.new(str,font,height,pos,color,align)
        # composite
        when "grid","paper","guides"
          composite = elem.get_elements("dia:composite")[0]
          val = parse_composite(composite)
        when "hguides","vguides"
          val = ""
        else
          return nil,nil,nil
        end
        return name,val
      end
    
      def parse_object(elem)
        info = {
          "type"       => elem.attribute("type").to_s,
          "version"    => elem.attribute("version").to_s,
          "id"         => elem.attribute("id").to_s,
          "attributes" => {}
        }
        elem.get_elements("dia:attribute").each do |attr|
          k,v = parse_attribute(attr)
          info["attributes"][k] = v if k
        end
        case info['type']
        when 'Standard - Arc'
          obj = Monpe::Items::Arc.new(info)
        when 'Standard - Box'
          obj = Monpe::Items::Box.new(info)
        when 'Standard - Ellipse'
          obj = Monpe::Items::Ellipse.new(info)
        when 'Embed - Image'
          obj = Monpe::Items::EmbedImage.new(info)
        when 'Embed - Text'
          obj = Monpe::Items::EmbedText.new(info)
        when 'Standard - Line'
          obj = Monpe::Items::Line.new(info)
        when 'Standard - Text'
          obj = Monpe::Items::Text.new(info)
        when 'ORCA - TextCircle'
          obj = Monpe::Items::TextCircle.new(info)
        else
          obj = nil
        end
        obj
      end

      def parse_group(elem)
        ret = []
        elem.elements.each do |child|
          case child.expanded_name
          when "dia:object"
            obj = parse_object(child)
            ret << obj if obj
          when "dia:group"
            g = parse_group(child)
            ret << g 
            ret
          end
        end
        ret
      end
    
      def parse_layer(elem)
        ret = {
          "name"    => elem.attribute("name").to_s,
          "visible"  => elem.attribute("visible").to_s == "true",
          "children" => []
        }
        elem.elements.each do |child| 
          case child.expanded_name
          when "dia:object"
            ret["children"] << parse_object(child)
          when "dia:group"
            g = parse_group(child)
            ret["children"] << g
          end
        end
        ret["children"].flatten!
        Monpe::Layer.new(ret["name"],ret["children"],ret["visible"])
      end

      def parse_dic_element(elem)
        name = elem.attribute("name").to_s
        ret = {
          "occurs" => elem.attribute("occurs").to_s.to_i
        }
        c =  elem.get_elements("appinfo/embed")
        c << elem.get_elements("dia:appinfo/dia:embed")
        c.flatten!
        if c.empty?
          ret["type"] = "node"
          ret["elements"] = {}
          children =  elem.get_elements("element")
          children << elem.get_elements("dia:element")
          children.flatten!
          children.each do |e|
            n,v = parse_dic_element(e)
            ret["elements"][n] = v
          end
        else
          ret["type"]   = c[0].attribute("object").to_s
          ret["length"] = c[0].attribute("length").to_s.to_i if ret["type"] == "text"
        end
        return name,ret
      end
    end
  end
end
