# coding:utf-8

require 'gdk_pixbuf2'
require 'pango'
require 'pp'

module Monpe

  class RendererError < StandardError; end

  #PDFに描画を行う内部クラス
  #ユーザは使用しない。
  #
  class Renderer

    # cmと72dpi pointの比
    DPI_SCALE = 72.0 / 2.54

    #:call-seq:
    #  new(filename) -> self
    #
    #Rendererを作成する。
    #
    #* 引数 filename: 出力するPDFファイルのパス
    #* 戻り値: self
    def initialize(filename)
      # for windows printing
      # see http://stackoverflow.com/questions/32051023/pango-cairo-layout-path-method-does-not-respect-kerning
      cairo_font_map = Pango::CairoFontMap.create("freetype")
      Pango::CairoFontMap.set_default(cairo_font_map)

      @filename = filename
      @surface = nil
      @context = nil
      @dash_length = 1.0
      @mode = Monpe::LINESTYLE_SOLID
    end

    #:call-seq:
    #  export(reports) -> nil
    #
    #Reportオブジェクトの配列を受け取りPDFファイルを作成する。
    #
    #* 引数 reports: Monpe::ReportのArray
    #* 戻り値: nil
    def export(reports)
      reports.each do |r|
        r.render(self)
      end
      finish
    end

    #--
    #以下はライブラリ内部でしか使用しないメソッド
    #++

    #:nodoc:
    def check
      raise RendererError.new("cairo surface is nil") unless @surface
      raise RendererError.new("cairo context is nil") unless @context
    end

    #:nodoc:
    def ensure_minimum_one_device_unit(value)
      ax,ay = @context.device_to_user_distance(1.0,1.0)
      return [ax,ay,value].max
    end

    #:nodoc:
    def begin_render(w,h,info = {})
      @xoffset = (info["xoffset"] || 0.0) * DPI_SCALE
      @yoffset = (info["yoffset"] || 0.0) * DPI_SCALE
      @left    = (info["lmargin"] || 0.0) * DPI_SCALE
      @top     = (info["tmargin"] || 0.0) * DPI_SCALE
      @right   = (info["rmargin"] || 0.0) * DPI_SCALE
      @bottom  = (info["bmargin"] || 0.0) * DPI_SCALE
      @r_scale = (info["scaling"] || 1.0) * DPI_SCALE
      @width   = w                        * DPI_SCALE
      @height  = h                        * DPI_SCALE
      #page setup
      if @surface
        @context.show_page
        @surface.set_size(@width,@height)
      else
        @surface = Cairo::PDFSurface.new(@filename,@width,@height)
        @context = Cairo::Context.new(@surface)
      end
      @context.translate(@left + @xoffset,@top + @yoffset)
      @context.scale(@r_scale,@r_scale)
    end

    #:nodoc:
    def end_render
      raise RendererError.new("@surface is nil") unless @surface
      @context.scale(1.0/@r_scale,1.0/@r_scale)
      @context.translate(-@left-@xoffset,-@top-@yoffset)

      # fill margin
      ul = Monpe::Point.new(0,0)
      lr = Monpe::Point.new(@left,@height)
      self.fill_rect(ul,lr,Monpe::Color.new(1,1,1))

      ul = Monpe::Point.new(@width - @right,0)
      lr = Monpe::Point.new(@width,@height)
      self.fill_rect(ul,lr,Monpe::Color.new(1,1,1))

      ul = Monpe::Point.new(0,0)
      lr = Monpe::Point.new(@width,@top)
      self.fill_rect(ul,lr,Monpe::Color.new(1,1,1))

      ul = Monpe::Point.new(0,@height - @bottom)
      lr = Monpe::Point.new(@width,@height)
      self.fill_rect(ul,lr,Monpe::Color.new(1,1,1))
    end

    #:nodoc:
    def finish
      raise RendererError.new("@surface is nil") unless @surface
      @surface.finish
      @context.destroy
    end

    #--
    #テキスト周り
    #++
    #:nodoc:
    def set_font(font,height)
      check

      @font = font
      font.calc_absolute_size(height)
      pfd = font.get_pango_font_description
      @layout = @context.create_pango_layout
      @layout.set_font_description(pfd)
    end

    #:nodoc
    TAKAO_SCALE = 0.0009765625
    #:nodoc
    ROSAI_SCALE = 0.0013077870775969962
    #:nodoc
    OCRB_SCALE  = 0.0013077870775969962

    #:nodoc:
    def draw_string(text,pos,alignment,color)
      return unless text
      return if text.size <= 0

      case @font.family
      when "TakaoMincho",
        "TakaoGothic",
        "TakaoPMincho",
        "TakaoPGothic",
        "TakaoExMincho",
        "TakaoExGothic"
        draw_string_widhout_pango(text,pos,alignment,color,TAKAO_SCALE)
        return
      when "OCRROSAI"
        draw_string_widhout_pango(text,pos,alignment,color,ROSAI_SCALE)
        return
      when "OCRB"
        draw_string_widhout_pango(text,pos,alignment,color,OCRB_SCALE)
        return
      end

      @context.set_source_rgb(color.red, color.green, color.blue)
      @context.save do |cr|
        @layout.set_alignment(Monpe.get_pango_alignment(alignment))
        @layout.set_text(text)
        iter = @layout.iter
        bline = iter.baseline
        extents = iter.line_extents[1] #ink_rect,logical_rect
        shift = 0.0
        case alignment
        when Monpe::ALIGN_CENTER
          shift = extents.rbearing / 2.0
        when Monpe::ALIGN_RIGHT
          shift = extents.width
        end
        mx = pos.x - shift * 1.0 / Pango::SCALE
        my = pos.y - bline * 1.0 / Pango::SCALE
        cr.move_to(mx,my)
        @layout.context_changed
        cr.show_pango_layout(@layout)
      end
    end

    #:nodoc:
    def draw_char(text,pos,color)
      return unless text
      return if text.size <= 0
      @context.set_source_rgb(color.red, color.green, color.blue)
      @context.save do |cr|
        @layout.set_text(text)
        iter = @layout.iter
        bline = iter.baseline
        mx = pos.x
        my = pos.y - bline * 1.0 / Pango::SCALE
        cr.move_to(mx,my)
        @layout.context_changed
        cr.show_pango_layout(@layout)
      end
    end

    #:nodoc
    def draw_string_widhout_pango(text,_pos,alignment,color,scale)
      return unless text
      return if text.size <= 0

      bb = calc_string_bb(text,_pos,alignment)
      pos = Monpe::Point.new(bb.left,_pos.y)
      ab_size = @layout.font_description.size
      text.split("").each do |chr|
        draw_char(chr,pos,color)
        if chr.ascii_only? or chr == "\u2212" #zenkaku minus
          pos.x += ab_size * scale / 2.0
        else
          pos.x += ab_size * scale
        end
      end
    end

    #:nodoc:
    def calc_string_bb(text,pos,alignment)
      bb = Monpe::Rectangle.new(pos.x,pos.y,pos.x,pos.y)
      return bb unless text
      return bb if text.size <= 0

      @context.save do |cr|
        @layout.set_alignment(Monpe.get_pango_alignment(alignment))
        @layout.set_text(text)
        iter = @layout.iter
        bline = iter.baseline
        extents = iter.line_extents[1] #ink_rect,logical_rect
        shift = 0.0
        case alignment
        when Monpe::ALIGN_CENTER
          shift = extents.rbearing / 2.0
        when Monpe::ALIGN_RIGHT
          shift = extents.width
        end
        top    = pos.y - ( 1.0 * bline + extents.ascent ) / Pango::SCALE
        bottom = pos.y + (-1.0 * bline + extents.descent) / Pango::SCALE
        left   = pos.x - (shift * 1.0                   ) / Pango::SCALE
        right  = pos.x + (extents.width - shift * 1.0   ) / Pango::SCALE

        scale = nil
        case @font.family
        when "TakaoMincho",
          "TakaoGothic",
          "TakaoPMincho",
          "TakaoPGothic",
          "TakaoExMincho",
          "TakaoExGothic"
          scale = TAKAO_SCALE
        when "OCRROSAI"
          scale = ROSAI_SCALE
        when "OCRB"
          scale = OCRB_SCALE
        end
        if scale
          ab_size = @layout.font_description.size
          w = 0.0
          text.split("").each do |chr|
            if chr.ascii_only? or chr == "\u2212" #zenkaku minus
              w += ab_size * scale / 2.0
            else
              w += ab_size * scale
            end
          end

          case alignment
          when Monpe::ALIGN_LEFT
            left = pos.x
          when Monpe::ALIGN_CENTER
            left = pos.x - w / 2.0
          when Monpe::ALIGN_RIGHT
            left = pos.x - w
          end
          right = left + w
        end
        bb = Monpe::Rectangle.new(left,top,right,bottom)
      end
      bb
    end

    #:nodoc:
    def draw_text(text)
      pos = text.get_position
      text.lines.each do |l|
        draw_text_line(l, pos, text.alignment, text.color)
        pos.y += text.height * 0.8
      end
    end

    #:nodoc:
    def draw_text_line(text_line, pos, align, color)
      set_font(text_line.font,text_line.height)
      draw_string(text_line.text,pos,align,color)
    end

    #:nodoc:
    def calc_text_line_bb(text_line, pos, align)
      set_font(text_line.font,text_line.height)
      calc_string_bb(text_line.text,pos,align)
    end

    #:nodoc:
    def calc_text_bb(text)
      pos = text.get_position
      bb = Rectangle.new(pos.x,pos.y,pos.x,pos.y)
      text.lines.each do |l|
        tl_bb = calc_text_line_bb(l,pos,text.alignment)
        pos.y += text.height * 0.8
        bb.union(tl_bb)
      end
      bb
    end

    #:nodoc:
    def get_width_pixels
      0
    end

    #:nodoc:
    def get_height_pixels
      0
    end

    #:nodoc:
    def get_text_width
      0
    end

    #:nodoc:
    def set_linewidth(line_width)
      if (line_width < 0.01)
        @context.set_line_width(2.54/400)
      else
        @context.set_line_width(line_width)
      end
    end

    #:nodoc:
    def set_linecaps(mode)
      case mode
      when Monpe::LINECAPS_BUTT
        @context.set_line_cap(Cairo::LineCap::BUTT)
      when Monpe::LINECAPS_ROUND
        @context.set_line_cap(Cairo::LineCap::ROUND)
      when Monpe::LINECAPS_PROJECTING
        @context.set_line_cap(Cairo::LineCap::SQUARE)
      else
        raise RendererError.new("Unsupported caps mode specified!")
      end
    end

    #:nodoc:
    def set_linejoin(mode)
      case mode
      when Monpe::LINEJOIN_MITER
        @context.set_line_join(Cairo::LineJoin::MITER)
      when Monpe::LINEJOIN_ROUND
        @context.set_line_join(Cairo::LineJoin::ROUND)
      when Monpe::LINEJOIN_BEVEL
        @context.set_line_join(Cairo::LineJoin::BEVEL)
      else
        raise RendererError.new("Unsupported join mode specified!")
      end
    end

    #:nodoc:
    def set_linestyle(mode)
      @mode = mode
      case mode
      when Monpe::LINESTYLE_SOLID
        @context.set_dash([],0)
      when Monpe::LINESTYLE_DASHED
        dashes = [
          @dash_length,
          @dash_length,
        ]
        @context.set_dash(dashes,0)
      when Monpe::LINESTYLE_DASH_DOT
        dashes = [
          @dash_length * 1.0,
          @dash_length * 0.45,
          @dash_length * 0.1,
          @dash_length * 0.45,
        ]
        @context.set_dash(dashes,0)
      when Monpe::LINESTYLE_DASH_DOT_DOT
        dashes = [
          @dash_length * 1.0,
          @dash_length * (0.8 / 3),
          @dash_length * 0.1,
          @dash_length * (0.8 / 3),
          @dash_length * 0.1,
          @dash_length * (0.8 / 3),
        ]
        @context.set_dash(dashes,0)
      when Monpe::LINESTYLE_DOTTED
        dashes = [
          @dash_length * 0.1,
          @dash_length * 0.1,
        ]
        @context.set_dash(dashes,0)
      else
        raise RendererError.new("Unsupported line mode specified!")
      end
    end

    #:nodoc:
    def set_dashlength(length)
      @dash_length = ensure_minimum_one_device_unit(length)
      set_linestyle(@mode)
    end

    #:nodoc:
    def set_fillstyle(mode)
      case mode
      when Monpe::FILLSTYLE_SOLID
      else
        raise RendererError.new("Unsupported fill mode specified!")
      end
    end

    #:nodoc:
    def draw_line(start_point, end_point, color)
      @context.set_source_rgb(color.red, color.green, color.blue)
      @context.move_to(start_point.x,start_point.y)
      @context.line_to(end_point.x,end_point.y)
      @context.stroke
    end

    #:nodoc:
    def draw_line_with_arrows(_start, _end, line_width, color, start_arrow, end_arrow)
      start_point = _start.dup
      end_point   = _end.dup
      start_arrow_head = start_point.dup
      if start_arrow and !start_arrow.is_none?
        move_arrow, move_line = start_arrow.calculate_arrow_point(start_point,end_point,line_width)
        start_arrow_head.sub(move_arrow)
        start_point.sub(move_line)
      end
      end_arrow_head = end_point.dup
      if end_arrow and !end_arrow.is_none?
        move_arrow, move_line = end_arrow.calculate_arrow_point(end_point,start_point,line_width)
        end_arrow_head.sub(move_arrow)
        end_point.sub(move_line)
      end
      draw_line(start_point,end_point,color)
      if start_arrow and !start_arrow.is_none?
        start_arrow.draw(self,start_arrow_head,end_point,line_width,color)
      end
      if end_arrow and !end_arrow.is_none?
        end_arrow.draw(self,end_arrow_head,start_point,line_width,color)
      end
    end

    #:nodoc:
    def _rect(ul,lr,color,fill)
      @context.set_source_rgb(color.red, color.green, color.blue)
      @context.rectangle(ul.x, ul.y, (lr.x - ul.x), (lr.y - ul.y))
      if fill
        @context.fill
      else
        @context.stroke
      end
    end

    #:nodoc:
    def draw_rect(ul,lr,color)
      _rect(ul,lr,color,false)
    end

    #:nodoc:
    def fill_rect(ul,lr,color)
      _rect(ul,lr,color,true)
    end

    #:nodoc:
    def draw_arc(center,width,height,angle1,angle2,color)
      @context.set_source_rgb(color.red, color.green, color.blue)
      @context.new_path
      start_x = center.x + width  / 2.0 * Math.cos((Math::PI/180.0) * angle1)
      start_y = center.y + height / 2.0 * Math.sin((Math::PI/180.0) * angle1)
      #@context.move_to(start_x,start_y)
      a1 = -1.0 * (angle1 / 180.0) * Math::PI
      a2 = -1.0 * (angle2 / 180.0) * Math::PI
      r2 = width > height ? height / 2.0 : width / 2.0
      @context.arc_negative(center.x,center.y,r2,a1,a2)
      @context.stroke
    end

    #:nodoc:
    def fill_arc(center,width,height,angle1,angle2,color)
      @context.set_source_rgb(color.red, color.green, color.blue)
      @context.new_path
      @context.move_to(center.x,center.y)
      @context.line_to(start_x,start_y)
      a1 = -1.0 * (angle1 / 180.0) * Math::PI
      a2 = -1.0 * (angle2 / 180.0) * Math::PI
      r2 = width > height ? height / 2.0 : width / 2.0
      @context.arc_negative(center.x,center.y,r2,a1,a2)
      @context.line_to(center.x,center.y)
      @context.close_path
      @context.fill
    end

    #:nodoc:
    def _ellipse(center,width,height,color,fill)
      @context.set_source_rgb(color.red, color.green, color.blue)
      @context.save do |cr|
        cr.new_sub_path
        xs = width*1.0/2.0
        ys = height*1.0/2.0
        cr.translate(center.x,center.y)
        cr.scale(xs,ys)
        cr.arc(0.0,0.0,1.0,0.0,2.0*Math::PI)
      end
      if fill
        @context.fill
      else
        @context.stroke
      end
    end

    #:nodoc:
    def draw_ellipse(center,width,height,color)
      _ellipse(center,width,height,color,false)
    end

    #:nodoc:
    def fill_ellipse(center,width,height,color)
      _ellipse(center,width,height,color,true)
    end

    #:nodoc:
    def draw_image(point,width,height,image)
      @context.save do |cr|
        cr.translate(point.x,point.y)
        cr.scale(width*1.0/image.width,height*1.0/image.height)
        surface = image.create_surface
        pattern = Cairo::SurfacePattern.new(surface)
        cr.set_source(pattern)
        cr.paint
        surface.finish
      end
    end

    def _rounded_rect(tl,br,color,radius,fill)
      radius = [radius,((br.x-tl.x)/2.0)].min
      radius = [radius,((br.y-tl.y)/2.0)].min
      @context.set_source_rgb(color.red, color.green, color.blue)
      @context.new_path

      # north-west
      @context.move_to((tl.x + radius),tl.y)
      # north-east
      @context.line_to((br.x - radius),tl.y)
      @context.arc((br.x - radius),(tl.y + radius),radius,-1.0 * Math::PI/2.0, 0)
      # soutch-east
      @context.line_to(br.x,(br.y - radius))
      @context.arc((br.x - radius),(br.y - radius),radius,0,Math::PI/2.0)
      # soutch-west
      @context.line_to(tl.x+radius,br.y)
      @context.arc((tl.x + radius),(br.y - radius),radius,Math::PI/2.0,Math::PI)
      # north-west
      @context.line_to(tl.x,tl.y + radius)
      @context.arc((tl.x + radius),(tl.y + radius),radius,Math::PI,-1.0 * Math::PI/2.0)

      if fill
        @context.fill
      else
        @context.stroke
      end
    end

    #:nodoc:
    def draw_rounded_rect(tl,br,color,radius)
      _rounded_rect(tl,br,color,radius,false)
    end

    #:nodoc:
    def fill_rounded_rect(tl,br,color,radius)
      _rounded_rect(tl,br,color,radius,true)
    end

    def _polygon(points,color,fill)
      if points.size < 1
        raise RendererError.new("_polygon:invalid points size")
      end
      @context.set_source_rgb(color.red, color.green, color.blue)
      @context.new_path
      @context.move_to(points[0].x,points[0].y)
      1.upto(points.size - 1) do |i|
        @context.line_to(points[i].x,points[i].y)
      end
      @context.line_to(points[0].x,points[0].y)
      @context.close_path
      if fill
        @context.fill
      else
        @context.stroke
      end
    end

    #:nodoc:
    def draw_polygon(points,color)
      _polygon(points,color,false)
    end

    #:nodoc:
    def fill_polygon(points,color)
      _polygon(points,color,true)
    end


    #:nodoc:
    def draw_arc_with_arrows(_start,_end,_mid,line_width,color,start_arrow,end_arrow)
      STDERR.puts "#{__method__} not implemented"
    end

    #:nodoc:
    def draw_rounded_polyline
      STDERR.puts "#{__method__} not implemented"
    end

    #:nodoc:
    def draw_polyline_with_arrows
      STDERR.puts "#{__method__} not implemented"
    end

    #:nodoc:
    def draw_rounded_polyline_with_arrows
      STDERR.puts "#{__method__} not implemented"
    end

    #:nodoc:
    def draw_bezier_with_arrows
      STDERR.puts "#{__method__} not implemented"
    end

    #:nodoc:
    def draw_bezier
      STDERR.puts "#{__method__} not implemented"
    end

    #:nodoc:
    def fill_bezier
      STDERR.puts "#{__method__} not implemented"
    end

    #:nodoc:
    def draw_polyline
      STDERR.puts "#{__method__} not implemented"
    end

    #:nodoc:
    def draw_object
      STDERR.puts "#{__method__} not implemented"
    end

    private

  end
end
