# coding:utf-8

module Monpe

  # monpeの帳票ファイル(*.red)に対応するクラス
  #
  class Report

    #レイヤーの配列
    attr_reader :layers

    #:call-seq:
    #  new(filename) -> self
    #
    #引数の帳票ファイルを読み込みReportオブジェクトを作成する。
    #
    #* 引数 filename: 帳票のパス
    #* 戻り値: self
    def initialize(filename)
      @info = Monpe::Parser.parse(filename)
      @dic_data   = @info["dictionarydata"]
      @layers     = @info["layers"]
      @paper_attr = @info["diagramdata"]["paper"]["val"]["attributes"]

      @paper_name    = @paper_attr["name"]
      @custom_width  = @paper_attr["custom_width"]
      @sustom_height = @paper_attr["custom_height"]
      @is_portrait   = @paper_attr["is_portrait"]

      @width, @height = Monpe::Paper.get_paper_size(@paper_name)
      @width, @height = @height, @width unless @is_portrait
      if @paper_name == "custom"
        @width  = @custom_width
        @height = @custom_height
      end

      @embed_objects = {}
      @layers.each do |layer|
        layer.items.each do |item|
          if item.is_a?(Monpe::Items::EmbedItem)
            @embed_objects[item.get_embed_id] = item
          end
        end
      end
    end

    #:call-seq:
    #  set_offset(x,y) -> nil
    #
    #オフセットを指定する
    #
    #* 引数 x: x方向のオフセット(cm)
    #* 引数 y: y方向のオフセット(cm)
    #* 戻り値: nil
    def set_offset(x,y)
      @paper_attr['xoffset'] = x
      @paper_attr['yoffset'] = y
      nil
    end 


    #:call-seq:
    #  render(renderer) -> nil
    #
    #帳票のレンダリングを行う。
    #通常ユーザは使用する必要がない。
    #
    #* 引数 renderer: Rendererオブジェクト
    #* 戻り値: nil
    def render(rdrr)
      rdrr.begin_render(@width,@height,@paper_attr)
      @layers.each do |layer|
        layer.render(rdrr) if layer.visible
      end
      rdrr.end_render
      nil
    end 

    #:call-seq:
    #  set_layer_visible(layer_name,visible) -> nil
    #
    #指定されたレイヤーの表示設定をする
    #
    #* 引数 layer_name: レイヤー名
    #* 引数 visible: 表示設定
    #* 戻り値: nil
    def set_layer_visible(layer_name,visible)
      @layers.each do |layer|
        if layer.name == layer_name
          layer.visible = visible
        end
      end
      nil
    end

    #:call-seq:
    #  set_all_layers_visible(visible) -> nil
    #
    #全レイヤーの表示設定をする
    #
    #* 引数 visible: 表示設定
    #* 戻り値: nil
    def set_all_layers_visible(visible)
      @layers.each do |layer|
        layer.visible = visible
      end
      nil
    end

    #:call-seq:
    #  get_embed_object(id) -> Monpe::Object::Object
    #
    #指定されたIDの埋め込みオブジェクトを返す。
    #
    #* 引数 id: 埋め込みオブジェクトのID
    #* 戻り値: 埋め込みオブジェクト、無効なidの場合はnil
    def get_embed_object(id)
      @embed_objects[id]
    end

    #:call-seq:
    #  get_embed_objects -> Hash
    #
    #埋め込みIDと埋め込みオブジェクトのハッシュを返す。
    #
    #* 戻り値: 埋め込みIDと埋め込みオブジェクトのハッシュ
    def get_embed_objects
      @embed_objects
    end
    
    #:call-seq:
    #  get_embed_data -> Hash
    #
    #帳票の埋め込みデータのテンプレートのHash(ID=>値)を返す。
    #Hashの値は全部空文字列に初期化される。
    #
    #* 戻り値: 埋め込みデータのテンプレートのHash
    def get_embed_data
      ret = {}
      @embed_objects.each do |k,v|
        ret[k] = ""
      end
      ret
    end

    #:call-seq:
    #  get_embed_data_orig -> Hash
    #
    #帳票の埋め込みデータのテンプレートのHash(ID=>値)を返す。
    #
    #* 戻り値: 埋め込みデータのテンプレートのHash
    def get_embed_data_orig
      ret = {}
      @embed_objects.each do |k,v|
        ret[k] = v.get_data
      end
      ret
    end

    #:call-seq:
    # set_embed_data(data) -> self
    #
    #埋め込みデータのHashから各埋め込みオブジェクトの埋め込み値を更新する。
    #
    #* 引数 data: get_embed_dataで取得した埋め込みデータのHash
    #* 戻り値: nil
    def set_embed_data(data)
      data.each do |k,v|
        v ||= ""
        @embed_objects[k].set_data(v) if @embed_objects[k]
      end
      nil
    end

    private

  end
end
